import os
import platform
import sys
import getpass
import re
from importlib import import_module


from tklib.tkobject import tkObject
from tklib.tkparams import tkParams
from tklib.tkutils import terminate, getarg, getintarg, getfloatarg, pint, pfloat, pconv_by_type
from tklib.tkutils import convert_by_vars, save_csv, del_quote, split_file_path

#=========================
# Application base class
#=========================


class tkApplication(tkObject):
    def __init__(self, usage_str = "", _globals = None, locals = None, 
                 use_user_inifile = False, inifile_dirs = ['ini'], create_inidir = False,
                 **args):
        super().__init__()

        self.os_name          = platform.system()
        self.args_opt_inf     = {}
        self.args_idx_inf     = []
        self.args_vars        = None

        self.set_usage(usage_str)
   
        self.phrases     = {}

        self.user            = getpass.getuser()
        self.user_nospace    = self.user.replace(' ', '')

        self.argv            = sys.argv
        self.script_path     = sys.argv[0]
        self.script_fullpath = os.path.abspath(self.script_path)
        dirname, basename, filebody, ext = split_file_path(self.script_fullpath)
        self.script_dir      = dirname
        self.script_filebody = filebody
#        self.script_dir      = os.path.splitext(self.script_fullpath)[0]

        self.ini_dir = self.get_inifile_dir(use_user_inifile = use_user_inifile,
                             inifile_dirs = inifile_dirs, create_inidir = create_inidir,
                              defval = self.script_dir)
        self.inifile = self.get_inifile_path(self.ini_dir, use_user_inifile = use_user_inifile)

        self.globals     = globals
        self.locals      = locals
        self.modules     = []


# Configuration parameters
        self.configparams   = tkParams()
        self.config         = self.configparams
        self.config.lang    = 'en'
        self.config.inifile = self.inifile
        self.config.ini_dir = os.path.dirname(os.path.abspath(self.config.inifile))

# Application global variables
        self.params = tkParams()

# Script variables
        self.svars  = tkParams()

        self.update(**args)

    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()

    def get_inifile_dir(self, use_user_inifile = False,
                        inifile_dirs = ['ini'], create_inidir = False,
                        defval = None):
        if not use_user_inifile:
            return self.script_dir

        for dir in inifile_dirs:
            path = os.path.join(self.script_dir, dir)
            path = os.path.abspath(path)
            if create_inidir or os.path.isdir(path):
                return path

        return defval

    def get_inifile_path(self, inidir, use_user_inifile = False):
        dirname, basename, filebody, ext = split_file_path(self.script_fullpath)
        if not use_user_inifile:
            inifile = os.path.join(inidir, f'{filebody}.ini')
        else:
            inifile = os.path.join(inidir, f'{filebody}_{self.user_nospace}.ini')
        
        inifile = os.path.abspath(inifile)
        
        return inifile

    def add_argument(self, opt, opt_str, var_name = None, desc = '', type = 'str', defval = None, optional = True):
        inf_opt = self.args_opt_inf
        inf_idx = self.args_idx_inf
        if opt is None:
            i = len(inf_idx)
            inf_idx.append( { "opt": i, "opt_str": opt_str, "var_name": var_name, "desc": desc, 
                              "type": type, "defval": defval, "optional": optional } )
        else:
            inf_opt[opt] = { "opt": opt, "opt_str": opt_str, "var_name": var_name, "desc": desc, 
                              "type": type, "defval": defval, "optional": optional }
        return inf_opt, inf_idx

    def read_args(self, vars = None, check_allowed_args = False):
        argv = sys.argv
        args_opt = {}
        args_idx = []
#        print(f"{vars=}")
        if vars is None:
            args_vars = {}
        else:
            args_vars = vars
            
        allowed_keys = self.args_opt_inf.keys()
        i_idx_arg = 0
        for i in range(1, len(argv)):
            arg = del_quote(argv[i])
            m = re.match(r'\s*(\S.*)\s*=\s*(.*)\s*$', arg)
            if m:
                opt = m.groups()[0]
                val = m.groups()[1]
                if check_allowed_args and opt not in allowed_keys:
                    return None, 1, f"Error in tkapplicaiton.read_args(): Invalid argument [{opt}]"
#                    print(f"Error in tkapplicaiton.read_args(): Invalid argument [{opt}]")
#                    exit()

                if opt in allowed_keys:
                    var_type = self.args_opt_inf[opt]["type"]
                    args_opt[opt] = pconv_by_type(val, var_type, defval = None, strict = True)
                    if args_opt[opt] is None:
                        return None, 2, \
                            f"Error: Argument [{opt}] = [{val}] is not compatible with the required type [{var_type}]"

                    var_name = self.args_opt_inf[opt]["var_name"]
                    if var_name is not None:
#                        print(f"set {var_name=} {args_opt[opt]=}")
                        args_vars[var_name] = args_opt[opt]
                else:
                    args_opt[opt] = val
            else:
                try:
                    var_type = self.args_idx_inf[i_idx_arg]["type"]
                    val = pconv_by_type(arg, var_type, defval = None, strict = True)
                    if val is None:
                        return None, 3, \
                            f"Error: #{i_idx_arg} positional argument = [{arg}] is not compatible with the required type [{var_type}]"

                    args_idx.append(val)
                    var_name = self.args_idx_inf[i_idx_arg]["var_name"]
                    if var_name is not None:
#                        print(f"set {var_name=} {args_opt[opt]=}")
                        args_vars[var_name] = val
                except:
                    args_idx.append(arg)

                i_idx_arg += 1

        for key in allowed_keys:
            v = self.args_opt_inf[key]
            var = v["var_name"]
            if args_opt.get(key, None) is None:
#                print(f"default {var=} {v['defval']=}")
                args_opt[var] = v["defval"]
                args_vars[var] = v["defval"]

                if not v["optional"]:
                    return None, 4, f"Error: Argument [{key}] is required but not given"
#                print("")
#                print(f"Error: Argument [{key}] is required but not given")
#                exit()

        nposargs = len(args_idx)
        for i in range(len(self.args_idx_inf)):
            v = self.args_idx_inf[i]
            if i >= nposargs:
                var = self.args_idx_inf[i]["var_name"]
                args_vars[var] = v["defval"]
                if not v["optional"]:
                    return None, 5, f"Error: #{i} positional argument is required but not given"
#                print("")
#                print(f"Error: #{i} positional argument is required but not given")
#                exit()
        
        self.args_opt  = args_opt
        self.args_idx  = args_idx
        self.args_vars = args_vars
#        print(f"{args_vars=}")
#        exit()
        
        return args_opt, args_idx, args_vars
    
    def get_arg(self, key, defval = None):
        if type(key) is int:
            try:
                return self.args_idx[key]
            except:
                return defval
        else:
            try:
                return self.args_opt[key]
            except:
                return defval

    def set_usage(self, usage_str = ''):
        self.usage_str   = usage_str

    def usage(self, *args):
        if self.usage_str is not None:
            if self.args_vars:
                s_conv = convert_by_vars(self.usage_str, self.args_vars, prefix = '', bra = r'\{', ket = r'\}')
            else:
                s_conv = self.usage_str

            for s in s_conv.split('\n'):
                if not re.match(r'\s*[\"\'].*[\"\']', s):
                    s = '"' + s + '"'
                cmd = 'print({})'.format(s.rstrip())
                eval(cmd)
            return

        inf_opt = self.args_opt_inf
        keys_opt = inf_opt.keys()
        inf_idx = self.args_idx_inf
#        print("inf_opt=", inf_opt)
#        print("inf_idx=", inf_idx)
        if len(keys_opt) > 0 or len(inf_idx) > 0:
            print("")
            usage = f'Usage: python {sys.argv[0]} '
            for key in keys_opt:
                v = inf_opt[key]
                if v['optional']:
                    usage += f"[{v['opt_str']}] "
                else:
                    usage += f"{v['opt_str']} "
            for i in range(len(inf_idx)):
                v = inf_idx[i]
                if v['optional']:
                    usage += f"[{v['opt_str']}] "
                else:
                    usage += f"{v['opt_str']} "

            print(usage)

            print(f"  Options:")
            for key in keys_opt:
                v = inf_opt[key]
                print(f"    {v['opt_str']} (def={v['defval']}) (type={v['type']})")
                if v['desc'] != '':
                    print(f"        {v['desc']}")
            for i in range(len(inf_idx)):
                v = inf_idx[i]
                print(f"    #{i} {v['opt_str']} (def={v['defval']}) (type={v['type']})")
                if v['desc'] != '':
                    print(f"        {v['desc']}")

            return

    def terminate(self, message = None, usage = None):
        if message is not None:
            print("")
            print(message)

        if usage:
            print("")
            usage(self)

        print("")

        exit()

    def write_resource(app, path, mode):
        with open(f, 'w') as fp:
            for key in app.phrases.keys():
                try:
                    val = app.phrases[key][lang]
                    fp.write(f'{key}:{val}' + '\n')
                except:
                    pass
        return True

    def read_resource(app, path, lang, phrases = None, sep = '='):
        if phrases is None:
            phrases = {}

        fp = open(path, 'r')
        if not fp:
            return phrases

        while 1:
            line = fp.readline()
            if not line:
                break

#            m = re.match(rf'\s*(\S.*)\s*=\s*(.*)\s*\n$', line, flags = re.MULTILINE | re.DOTALL)
            m = re.match(rf'\s*(\S.*)\s*{sep}\s*(.*)\s*\n$', line, flags = re.MULTILINE | re.DOTALL)
            if not m:
                continue
            
            var = m.groups()[0]
            val = m.groups()[1]
            if phrases.get(var, None) is None:
                phrases[var] = {}
            
            phrases[var][lang] = val
#            print(f"{var=} {val=}")

        fp.close()
#        print(f"{phrases=}")
#        exit()
        
        return phrases

    def p(self, str, defval = None, phrases = None):
        if phrases is None:
            phrases = self.phrases

        if str is None or str == '':
            return ''

#        print("")
#        print(f"{str=}")
        s = phrases.get(str, defval)
#        print(f"{s=}")
        if s is None:
            if defval is not None:
                print(f" 336 return defval={defval}")
                return defval
 
#            print(f" return {str=}")
            return str

        s = s.get(self.config.lang, defval)
#        print(f"{s=}")
        if s is None:
            if defval is not None:
#                print(f" 346 return defval={defval}")
                return defval
 
#            print(f" return {str=}")
            return str

#        print(f"last: return {s=}")
        return s

    def print(self, *args, **kwargs):
        print(*args, **kwargs)

    def get_args(self):
        argv = sys.argv
        return argv, len(argv)

    def get_param_dic(self):
        return self.params.__dict__

    def print_dict(self, d, header = None):
        print("")
        if header is not None:
            print(header)
        
        if d is None:
            print("  Null dict")
            return

        for key in d.keys():
            print(f"{key}: {d[key]}")


    def load_module(self, module_name, desc = ''):
        try:
            module = import_module(module_name)
            self.modules.append({"module": module, "module_name": module_name, "desc": desc})
            return module
        except:
            return None

    def call(self, module_name, func_name, *args, **kwargs):
        ret = None
        for minf in self.modules:
            if minf["module_name"] == module_name:
#                print("module:", minf["module_name"], minf["module"])
                module = minf["module"]
#                print("module names:", module.__dict__.keys())
                
                if func_name not in module.__dict__.keys():
#                    return None
                    pass
                else:
                    ret = module.__dict__[func_name](*args, **kwargs)
                    return ret

#        print(f"call globals.{func_name}")
        if func_name not in self.globals.keys():
            self.terminate(f"Error in tkapplication.call(): Can not find the function [{func_name}] in self.globals\n")
#            print("")
#            print(f"Error in tkapplication.call(): Can not find the function [{func_name}] in self.globals\n")
#            return None
        else:
            ret = self.globals[func_name](*args)

        return ret

    def call_all(self, func_name, *args):
        ncall = 0
        ret = None
        for minf in self.modules:
#            print("module:", minf["module_name"], minf["module"])
            if minf["module"].__dict__.get(func_name, None) is not None:
                ret = minf["module"].__dict__[func_name](*args)
                ncall += 1

        if ncall == 0:
#            print(f"call globals.{func_name}")
            ret = self.globals[func_name](*args)

        return ret

    def get_argv(self):
        return sys.argv, len(sys.argv)

    def get_params(self):
        return self.params

    def get_param_dict(self):
        return self.params.get_param_dict()

    def read_parameters(self, path, merge_params = False, AddSection = 0):
        ini = tkIniFile()
        inf = ini.ReadAll(path, AddSection = AddSection)
        if inf is not None and merge_params:
            self.params += inf

        return inf

    def save_parameters(self, path, params = None, section = 'Parameters', otherparams = None):
        if params is None:
            params = self.get_params()

        ini = tkIniFile(path)
        for i in params.keys():
            val = params.get(key, None)
            if val is not None:
                ini.WriteString(section, key, val)

        if otherparams is not None:
            for key in otherparams.keys():
                val = otherparams.get(key, None)
                if val is not None:
                    ini.WriteString('Preferences', key, val)

    def print_parameters(self, params = None):
        if params is None:
            params = self.get_params()

        for key in params.keys():
            val = params.get(key, None)
            if val is None:
                print("  {:10s}: {:14.8g}".format(key, 'None'))
            else:
                print("  {:10s}: {:14.8g}".format(key, val))


def main():
    print("")
    print("This is library, not runnable")
    print("")


if __name__ == "__main__":
    main()
